/* Copyright (C) 2000-2006 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_IO_H
#include <io.h>
#endif
#include <errno.h>
#include <math.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_searchtool.h"
#include "udm_boolean.h"
#include "udm_vars.h"
#include "udm_conf.h"
#include "udm_log.h"


/********** QSORT functions *******************************/

static int cmpurlid (UDM_URL_CRD *s1, UDM_URL_CRD *s2)
{
  if (s1->url_id > s2->url_id) return(1);
  if (s1->url_id < s2->url_id) return(-1);
  if (s1->secno > s2->secno) return 1;
  if (s1->secno < s2->secno) return -1;
  return (int) s1->pos - (int) s2->pos;
}


static int
cmppattern(UDM_URLDATALIST *Data, UDM_URLDATA *D,
           long j, const char *pattern)
{
  int rc;

  for(; *pattern != '\0'; pattern++)
  {
    switch(*pattern)
    {
      case 'R':
      case 'r':
        if (D->coord > Data->Item[j].coord) return (*pattern == 'R') ? 1 : -1;
        if (D->coord < Data->Item[j].coord) return (*pattern == 'R') ? -1 : 1;
        break;
      case 'P':
      case 'p':
        if (D->pop_rank > Data->Item[j].pop_rank) return (*pattern == 'P') ? 1 : -1;
        if (D->pop_rank < Data->Item[j].pop_rank) return (*pattern == 'P') ? -1 : 1;
        break;
      case 'D':
      case 'd':
        if (D->last_mod_time > Data->Item[j].last_mod_time) return (*pattern == 'D') ? 1 : -1;
        if (D->last_mod_time < Data->Item[j].last_mod_time) return (*pattern == 'D') ? -1 : 1;
        break;
      case 'U':
      case 'u':
        rc= strcmp(UDM_NULL2EMPTY(D->url), UDM_NULL2EMPTY(Data->Item[j].url));
        if (rc) return(*pattern == 'U' ? -rc : rc);
        break;
      case 'S':
      case 's':
        rc= strcmp(UDM_NULL2EMPTY(D->section),
                   UDM_NULL2EMPTY(Data->Item[j].section));
        if (rc) return(*pattern == 'S' ? -rc : rc);
        break;
    }
  }
  return 0;
}


static int
cmppatternRP(UDM_URLDATA *D1, UDM_URLDATA *D2)
{
  if (D1->coord > D2->coord) return -1;
  if (D1->coord < D2->coord) return  1;
  
  if (D1->pop_rank > D2->pop_rank) return -1;
  if (D1->pop_rank < D2->pop_rank) return 1;
  return 0;
}


static int
cmppatternR(UDM_URLDATA *D1, UDM_URLDATA *D2)
{
  if (D1->coord > D2->coord) return -1;
  if (D1->coord < D2->coord) return  1;
  return 0;
}



/****************************************************/

void UdmSortSearchWordsByURL(UDM_URL_CRD *wrd,size_t num)
{
  if(wrd && num)UdmSort((void*)wrd,num,sizeof(*wrd),(udm_qsort_cmp)cmpurlid);
}


static int
cmpsiteid2(UDM_URLDATA *p1, UDM_URLDATA *p2)
{
  if (p1->site_id  > p2->site_id)  return -1;
  if (p1->site_id  < p2->site_id)  return 1;
#if 0
  if (p1->coord    > p2->coord)    return -1;
  if (p1->coord    < p2->coord)    return 1;
  if (p1->pop_rank > p2->pop_rank) return -1;
  if (p1->pop_rank < p2->pop_rank) return 1;
#endif
  return 0;
}


void
UdmURLDataSortBySite(UDM_URLDATALIST *L)
{
  UdmSort((void*) L->Item, L->nitems, sizeof(*L->Item), (udm_qsort_cmp) cmpsiteid2);
#if 0
  {
    size_t i;
    printf("L->num: %d num: %d\n", L->ncoords, num);
    for (i= 0; i < num; i++)
      printf("%d %d %d %d %.6f\n", i,
      L->Data[i].url_id, L->Data[i].site_id, L->Data[i].coord, L->Data[i].pop_rank);
  }
#endif
}

static void
UdmURLDataSortByPatternRP(UDM_URLDATALIST *L)
{
  UdmSort((void*) L->Item, L->nitems, sizeof(*L->Item), (udm_qsort_cmp) cmppatternRP);
}


static void
UdmURLDataSortByPatternR(UDM_URLDATALIST *L)
{
  UdmSort((void*) L->Item, L->nitems, sizeof(*L->Item), (udm_qsort_cmp) cmppatternR);
}


static size_t UdmH[] = {1, 5, 19, 41, 109, 209, 505, 929, 2161,
                        3905, 8929, 16001, 36289, 64769};
void UdmURLDataSortByPattern(UDM_URLDATALIST *L, const char *pattern)
{
  register ssize_t h, i, j;
  int s = 13;
  size_t num= L->nitems;
  UDM_URLDATA Dat;

  if (!strcmp(pattern, "RP"))
  {
    UdmURLDataSortByPatternRP(L);
    goto ret;
  }
  if (!strcmp(pattern, "R"))
  {
    UdmURLDataSortByPatternR(L);
    goto ret;
  }

  while( (s > 0) && ((num / 3) < UdmH[s])) s--;
  while(s >= 0)
  {
    h= UdmH[s];
    for (j= h; j < (ssize_t) num; j++)
    {
      Dat= L->Item[j];
      i= j - h;
D4:
      if (cmppattern(L, &Dat, i, pattern) <= 0) goto D6;
      L->Item[i + h]= L->Item[i];
      i-= h;
      if (i >= 0) goto D4;
D6:
      L->Item[i + h]= Dat;
    }
    s--;
  }

ret:

#if 0
  {
    size_t i;
    printf("L->num: %d num:%d pattern:%s\n", L->ncoords, num, pattern);
    for (i= 0; i < num; i++)
    {
      printf("[%d]%d %d %d %d %.6f\n", i,
             L->Coords[i].url_id, L->Coords[i].coord,
             L->Data[i].url_id, L->Data[i].coord, (float) L->Data[i].pop_rank);
    }
  }
#endif
  return;
}



static int
cmp_score(UDM_URL_SCORE *c1, UDM_URL_SCORE *c2)
{
  return ((int4) c2->score) - ((int4) c1->score);
}


static int
cmp_score_then_url_id(UDM_URL_SCORE *c1, UDM_URL_SCORE *c2)
{
  int diff= ((int4) c2->score) - ((int4) c1->score);
  if (diff)
    return diff;
  return c1->url_id - c2->url_id;
}


/* Find topcount best results */

#define UDM_URL_SCORE_IS_SMALLER(x,y) \
 (x->score < y->score ||(x->score == y->score && x->url_id >= y->url_id))

void
UdmURLScoreListSortByScoreThenURLTop(UDM_URLSCORELIST *ScoreList,
                                     size_t topcount)
{
  UDM_URL_SCORE *First= ScoreList->Item;
  UDM_URL_SCORE *Last= ScoreList->Item + ScoreList->nitems;
  UDM_URL_SCORE *TopCount= ScoreList->Item + topcount;
  UDM_URL_SCORE *Curr;

  UDM_ASSERT(topcount < ScoreList->nitems);
/*
  for (Curr= First; Curr < Last; Curr++)
    fprintf(stderr, "%d:%d\n", Curr->score, Curr->url_id);
*/  
  UdmSort((void*) ScoreList->Item, topcount + 1,
          sizeof(UDM_URL_SCORE), (udm_qsort_cmp) cmp_score_then_url_id);

  for (Curr= TopCount; Curr < Last; Curr++)
  {
    UDM_URL_SCORE tmp;
    UDM_URL_SCORE *Left, *Right;
    if (UDM_URL_SCORE_IS_SMALLER(Curr, TopCount))
        continue;

    /*
      The Curr item is bigger than the TopCount item.
      Find its sorted place, and insert into the "topcount" list.
      Put the old TopCount item instead of the Curr item.
    */
    for(Left= First, Right= TopCount; Left < Right; )
    {
      UDM_URL_SCORE *Middle= Left + (Right - Left) / 2;
      if (UDM_URL_SCORE_IS_SMALLER(Curr, Middle))
        Left= Middle + 1;
      else
        Right= Middle;
    }
    tmp= *TopCount;
    memmove(Right + 1, Right, (TopCount-Right)*sizeof(UDM_URL_SCORE));
    *Right= *Curr;
    *Curr= tmp;
  }
}


void UdmURLScoreListSortByScore(UDM_URLSCORELIST *ScoreList)
{
  UdmSort((void*) ScoreList->Item, ScoreList->nitems,
          sizeof(UDM_URL_SCORE), (udm_qsort_cmp) cmp_score);
}


void UdmURLScoreListSortByScoreThenURL(UDM_URLSCORELIST *ScoreList)
{
  UdmSort((void*) ScoreList->Item, ScoreList->nitems,
          sizeof(UDM_URL_SCORE), (udm_qsort_cmp) cmp_score_then_url_id);
}


#define NUMWORD_FACTOR_SIZE 256
#define NUMWORD_FACTOR(nwf, n) (nwf[n >= NUMWORD_FACTOR_SIZE ? NUMWORD_FACTOR_SIZE - 1 : n])

typedef struct udm_score_param_st
{
  unsigned int *R;
  unsigned int *D;
  size_t D_size;
  size_t ncosine;
  size_t nsections;
  float Rsum_factor;
  size_t dst_offs;
  size_t nwf_offs;
  size_t nwf_num;
  unsigned int dst_weight;
  UDM_WIDEWORD *Word;
  size_t nuniq;
  size_t nwords;
  int search_mode;
  size_t *count;
  size_t count_size;
  UDM_STACKITEMLIST ItemList;
  char wf[256];
  char wf2[256];
  char nwf[256];
  float numword_factor[NUMWORD_FACTOR_SIZE];
  float numdistinctword_factor[NUMWORD_FACTOR_SIZE];
  float max_coord_factor;
  unsigned int MinCoordFactor;
  unsigned int NumDistinctWordFactor;
  int have_WordFormFactor;
  float WordFormFactor;
  float WordFormFactorReminder;
  int SaveSectionSize;
  float WordDensityFactor;
  float WordDensityFactorReminder;
} UDM_SCORE_PARAM;



/*
  R[i] and D[i] are in the range 0..64.
  ns is between 1..256
*/


#define MAXCOORD_FACTOR(factor, max_coord) ((1-((float)factor*(float)max_coord)))
#define MINCOORD_FACTOR(factor, coord) ((float) 0x1000 / (float) (0x1000 + factor * (coord - 1)))


static inline float
UdmMinMaxCoordFactor(UDM_SCORE_PARAM *score_param,
                     UDM_URL_CRD *B, UDM_URL_CRD *E)
{
  float x;
  x= MAXCOORD_FACTOR(score_param->max_coord_factor, E->pos) *
     MINCOORD_FACTOR(score_param->MinCoordFactor, B->pos);
  /*
  fprintf(stderr, "[%d] secno=%d min=%d max=%d x=%.8f\n",
                  B->url_id, (int) B->secno, (int) B->pos, (int) E->pos, x);
  */
  return x;
}

typedef struct udm_coord_st
{
  unsigned short pos;
  unsigned short ord;
} UDM_COORD;


/* 
  All words must be from the same section
*/
static void
CalcAverageWordDistance(UDM_WIDEWORD *Word, char *wf2,
                        UDM_URL_CRD *phr, size_t num,
                        size_t *sum, size_t *np)
{
  UDM_COORD prev;
  UDM_COORD curr;
  UDM_COORD next;
  UDM_URL_CRD *last;
  size_t wf2_secno;

  if (num < 2)
    return;
  
  wf2_secno= wf2[phr->secno];
  
  if (num == 2)
  {
    size_t num0= phr[0].num;
    size_t num1= phr[1].num;
    size_t ord0= Word[num0].order;
    size_t ord1= Word[num1].order;
    size_t pos0= phr[0].pos;
    size_t pos1= phr[1].pos;
    size_t res= (ord0 == ord1) ? 0 : pos1 > pos0 ? pos1 - pos0 : pos0 - pos1;
    (*np)++;
    (*sum)+= res > 0 ? (res - 1) * wf2_secno : 0;
    return;
  }
  
  
  last= phr + num;
  
  prev.pos= phr->pos;
  prev.ord= Word[phr->num].order;
  phr++;
  
  curr.pos= phr->pos;
  curr.ord= Word[phr->num].order;
  phr++;
  
  for (; phr < last; phr++)
  {
    next.pos= phr->pos;
    next.ord= Word[phr->num].order;
    
    if (prev.ord == curr.ord)  
    {
      if (curr.ord == next.ord)
      {
        /* w1 w1 w1 */
      }
      else
      {
        /* w1 w1 w2 */
      }
    }
    else
    {
      if (curr.ord == next.ord)
      {
        /* w1 w2 w2 */
      }
      else
      {
        uint4 diff1, diff2;
        diff1= curr.pos - prev.pos;
        diff2= next.pos - curr.pos;
        if (diff1 > 0) diff1--;
        if (diff2 > 0) diff2--;
        if (prev.ord == next.ord)
        {
          /* w1 w2 w1 */
          (*sum)+= (diff1 < diff2 ? diff1 : diff2) * wf2_secno;
          (*np)++;
        }
        else
        {
          /* w1 w2 w3 */
          (*sum)+= (diff1 + diff2) * wf2_secno;
          (*np)+= 2;
        }
      }
    }
    prev= curr;
    curr= next;
  }
}



static uint4
UdmCalcCosineWeightQuick(UDM_SCORE_PARAM *score_param,
                         UDM_URL_CRD *Crd, size_t ncoords,
                         char *wf2, float nwords_factor)
{
  float res;

  if (ncoords == 1)
  {
    size_t wrdsec= Crd->secno;
    size_t wrdnum= score_param->Word[Crd->num].order;
    size_t offs= score_param->nsections * wrdnum + wrdsec;
    nwords_factor*= MAXCOORD_FACTOR(score_param->max_coord_factor, Crd->pos) *
                    MINCOORD_FACTOR(score_param->MinCoordFactor, Crd->pos);
    res= score_param->Rsum_factor * nwords_factor *
         (float) (score_param->R[offs]) + 0.5;
  }
  else
  {
    size_t Dsum=0, RDsum= 0;
    char *added= (char*) score_param->D;
    float min_max_factor= 1;
    
    bzero(added, score_param->ncosine);
  
    if (score_param->nwords == 1)
    {
      unsigned char secno_prev= Crd->secno;
      UDM_URL_CRD *Crd_prev= Crd;
      size_t weight= score_param->Word[0].weight;
      for( ; ncoords; Crd++, ncoords--)
      {
        size_t wrdsec= Crd->secno;

        if (secno_prev != wrdsec)
        {
          min_max_factor*= UdmMinMaxCoordFactor(score_param, Crd_prev, Crd-1);
          secno_prev= wrdsec;
          Crd_prev= Crd;
        }
        
        if (!added[wrdsec])
        {
          size_t add= wf2[wrdsec] + weight;
          Dsum+= add*add;
          RDsum+= add * score_param->R[wrdsec];
          added[wrdsec]= 1;
        }
      }
      min_max_factor*= UdmMinMaxCoordFactor(score_param, Crd_prev, Crd-1);
    }
    else
    {
      size_t dstnum= 0, dstsum= 0;
      unsigned char secno_prev= Crd->secno;
      UDM_URL_CRD *Crd_prev= Crd;

      for( ; ncoords; Crd++, ncoords--)
      {
        unsigned char wrdsec= Crd->secno;
        size_t wrdnum= score_param->Word[Crd->num].order;
        size_t offs= score_param->nsections * wrdnum + wrdsec;
        
        if (secno_prev != wrdsec)
        {
          min_max_factor*= UdmMinMaxCoordFactor(score_param, Crd_prev, Crd-1);
          if (score_param->nuniq > 1)
            CalcAverageWordDistance(score_param->Word, wf2, Crd_prev, Crd - Crd_prev, &dstsum, &dstnum);
          secno_prev= wrdsec;
          Crd_prev= Crd;
        }
        
        if (!added[offs])
        {
          size_t add= wf2[wrdsec] + score_param->Word[wrdnum].weight;
          Dsum+= add*add;
          RDsum+= add * score_param->R[offs];
          added[offs]= 1;
        }
      }
      min_max_factor*= UdmMinMaxCoordFactor(score_param, Crd_prev, Crd-1);
      if (score_param->nuniq > 1)
        CalcAverageWordDistance(score_param->Word, wf2, Crd_prev, Crd - Crd_prev, &dstsum, &dstnum);
      
      if (dstnum)
      {
        dstsum= dstsum * score_param->dst_weight / dstnum / 255;
        Dsum+= dstsum * dstsum;
        RDsum+= score_param->R[score_param->dst_offs] * dstsum;
      }
    }

    res= score_param->Rsum_factor * nwords_factor * min_max_factor *
          (float) RDsum / sqrt(Dsum) + 0.5;
  }
  return (uint4) res;
}



static int UdmOriginWeight(int origin)
{
  switch(origin)
  {
    case UDM_WORD_ORIGIN_QUERY: return 3;
    case UDM_WORD_ORIGIN_SPELL: return 1;
  }
  return 0;
}


static void UdmWideWordListSetOriginWeight(UDM_WIDEWORDLIST *WWList)
{
  size_t i;
  for (i=0; i < WWList->nwords; i++)
    WWList->Word[i].weight= UdmOriginWeight(WWList->Word[i].origin);
}


#if 0
static void
UdmPrintCoords(UDM_URL_CRD *Coords, size_t ncoords)
{
  size_t i;
  for (i=0 ; i < ncoords; i++)
    fprintf(stderr, "[%d] %d %d %d/%d %d\n", i,
                    Coords[i].url_id,
                    Coords[i].secno,
                    Coords[i].pos,
                    Coords[i].seclen,
                    Coords[i].num);
}
#endif

static inline void CheckPhrase(
       UDM_WIDEWORDLIST *WWL,
       UDM_STACK_ITEM *query, size_t nitems,
       UDM_URL_CRD *coords, size_t ncoords,
       size_t *count)
{
  size_t q;
  size_t start, end, arg, i;
  size_t rstart, rend;
  size_t j, d, delta;
  UDM_URL_CRD *prev;

  /* find opening phrase command */
  for (q= 0; q < nitems; q++)
  {
    int cmd_arg;
    if (query[q].cmd != UDM_STACK_PHRASE) continue;
    cmd_arg= query[q].arg;
    
    /* find closing phrase command */
    start= q + 1;
    for (end= start; end < nitems && query[end].cmd != UDM_STACK_PHRASE; end++);
    q= end;
    arg= 0;

    /* skip trailing stopwords for now */
    /* TODO: we have to check document length (for phrases like "word1 stopword1") */
    for (rstart= start; rstart < end && query[rstart].cmd == UDM_STACK_STOP; rstart++);
    for (rend= end; rend > rstart && query[rend].cmd == UDM_STACK_STOP; rend--);

    /* if phrase contains stopwords only, we assume this document is found */
    if (rstart == rend) arg= 1;
    else for (i= 0; i < ncoords; i++)
    {
      size_t ord_i= WWL->Word[coords[i].num].order;
      if (query[rstart].arg != ord_i) continue;
      prev = &coords[i];
      d= i + 1;
      delta= 1;

      for (j= rstart + 1; j < rend; j++)
      {
        if (query[j].cmd == UDM_STACK_STOP)
        {
          delta++;
          continue;
        }

        /* find coord for this word */
        while (d < ncoords &&
               (UDM_WRDDOCPOS(prev->secno, prev->pos) == UDM_WRDDOCPOS(coords[d].secno, coords[d].pos) ||
                (UDM_WRDDOCPOS(prev->secno, prev->pos) + delta == UDM_WRDDOCPOS(coords[d].secno,coords[d].pos) &&
                 WWL->Word[coords[d].num].order != query[j].arg))) d++;

        if (d == ncoords ||
            UDM_WRDDOCPOS(prev->secno,prev->pos) != UDM_WRDDOCPOS(coords[d].secno,coords[d].pos) - delta ||
            query[j].arg != WWL->Word[coords[d].num].order) break;

        delta= 1;
        prev= &coords[d];
      }
      if (j == rend)
      {
        if (cmd_arg == '=' && rend - rstart + 1 != coords[i].seclen)
          continue;
        
        arg= 1;
        break;
      }
    }

    for (i= rstart; i < rend; i++)
      if (query[i].cmd == UDM_STACK_WORD) count[query[i].arg]= arg;
  }
}

/*
  TODO: implement tf*idf as follows:
  w_ik = tf_ik * idf_k

  where

  T_k = term k in document D_i
  tf_ik = frequency of term T_k in document D_i
  idf_k = inversed document frequency of term T_k in the collection C:

  idf_k = log( N / n_k)

  N = total number of documents in the collection C (i.e. in the result)
  n_k = the number of documents in C that contain T_k

  QQ: how to combine several words and sections?
  QQ: "number of documents in C" should probably mean
      "number of documents in the current search result" in our case.
*/

/*
#define DEBUG_REL 1
*/

#if 0
static void UdmGroupByURLNewRel(UDM_RESULT *Res)
{
  urlid_t prev_id;
  size_t prev_num= 0;
  size_t N= 0;
  size_t M= Res->WWList.nuniq;
  size_t ndoc= 0; 
  size_t i;
  UDM_URL_CRD *Crd = Res->CoordList.Coords;
  float  TFi[UDM_MAXWORDPERQUERY+1];
  float   Ci[UDM_MAXWORDPERQUERY+1];
  float   Vi[UDM_MAXWORDPERQUERY+1];    
  float   Wi[UDM_MAXWORDPERQUERY+1];
  float   Di[UDM_MAXWORDPERQUERY+1];
    
  bzero(TFi, sizeof(TFi));
  bzero(Ci, sizeof(Ci));
  bzero(Vi, sizeof(Vi));
  bzero(Di, sizeof(Di));
  
  
  /* N - number of documents */
  for (prev_id= Crd[0].url_id, i=0; i<= Res->CoordList.ncoords; i++)
  {
    if ( (i==Res->CoordList.ncoords) || (prev_id != Crd[i].url_id))
    {
      N++;
      if (i < Res->CoordList.ncoords)
      {
        prev_id= Crd[i].url_id;
        prev_num= i;
      }
    }

#ifdef DEBUG_REL
    if (i < Res->CoordList.ncoords)
    {
      fprintf(stderr,"[%d]doc=%d sec=%d wrd=%d pos=%d\n",
              i,Crd[i].url_id,
              UDM_WRDSEC(Crd[i].coord),
              UDM_WRDNUM(Crd[i].coord),
              UDM_WRDPOS(Crd[i].coord));
    }
#endif
  }
    
    
  /* TFi */
  for (i=0; i < Res->CoordList.ncoords; i++)
  {
    TFi[UDM_WRDNUM(Crd[i].coord)]++;
  }
    
  /* Add fictious word */
  TFi[M]= ((float) Res->CoordList.ncoords) / M;
  M++;
    
    
    
  /* Ci */
  for (i=0; i < M; i++)
  {
     Ci[i]= ((float) TFi[i]) / N;
#ifdef DEBUG_REL
     fprintf(stderr,"[%d] TDi=%d Ci=%f\n",i,TFi[i], Ci[i]);
#endif
  }
    
    
  /* Vi */
  for (prev_id= Crd[0].url_id, i=0; i<= Res->CoordList.ncoords; i++)
  {
    size_t k;
      
    if ( (i==Res->CoordList.ncoords) || (prev_id != Crd[i].url_id))
    {
      Di[M-1]= TFi[M-1]/N; /* fictious word */
      for (k=0; k < M; k++)
      {
        Vi[k]+= (Di[k] - Ci[k]) * (Di[k] - Ci[k]);
      }
      bzero(Di, sizeof(Di));
        
      if (i < Res->CoordList.ncoords)
      {
        prev_id= Crd[i].url_id;
        prev_num= i;
      }
    }
    if (i < Res->CoordList.ncoords)
      Di[UDM_WRDNUM(Crd[i].coord)]++;
  }
    
    
  for (i=0; i < M; i++)
  {
    Vi[i]/= (N-1);
    Wi[i]= Ci[i] ? TFi[i]*Vi[i]/Ci[i]/Ci[i] : 0;
#ifdef DEBUG_REL
    fprintf(stderr,"[%d] Vi=%f Wi=%f\n", i, Vi[i],Wi[i]);
#endif
  }
    
    
  bzero(Di,sizeof(Di));
  for (prev_id= Crd[0].url_id, i=0; i<= Res->CoordList.ncoords; i++)
  {
    size_t k;
    
    if ( (i==Res->CoordList.ncoords) || (prev_id != Crd[i].url_id))
    {
      float sumWiDi= 0;
      float sumDi2= 0;
      float sumWi2= 0;
      float res;
      size_t n= 1;
        
      Di[M-1]= TFi[M-1]/N;  /* fictious word */
       
      for (k=0; k < M; k++)
      {
        sumWiDi+= Wi[k] * Di[k] / (float)M;
        sumDi2+=  Di[k] * Di[k] / (float)M;
        sumWi2+=  Wi[k] * Wi[k];
        if (k + 1 < M)
          n*= Di[k];
      }
        
      res= sumWiDi / sqrt(sumDi2 * sumWi2);
      Crd[ndoc].url_id= prev_id;
      Crd[ndoc].coord= res * 100000;
        
#ifdef DEBUG_REL
      fprintf(stderr,"RES[%d]=%f %f %f\n",
        prev_id, res, sumWiDi, sqrt(sumDi2 * sumWi2));
#endif 
        
      ndoc++;
        
      bzero(Di, sizeof(Di));
      if (i < Res->CoordList.ncoords)
      {
        prev_id= Crd[i].url_id;
        prev_num= i;
      }
    }
      
    if (i <= Res->CoordList.ncoords)
      Di[UDM_WRDNUM(Crd[i].coord)]++;
  }
    
  Res->CoordList.ncoords= ndoc;
  return;
  
}
#endif



static
void UdmNumWordFactorInit(float k, float *numword_factor, size_t ncoords)
{
  size_t i;
  if (k > 1) k= 1;
  if (k < 0) k= 0;
  for (i= 0; i < NUMWORD_FACTOR_SIZE; i++)
  {
    numword_factor[i]= (1-k) + ((float) i / ncoords)*k;
  }
}


#define UDM_MIX_FACTOR(Factor, newfactor, Reminder, oldfactor) \
((Factor) * (newfactor) + (Reminder) * (oldfactor))

static inline void
UdmCalcWordDensityFactor(UDM_SCORE_PARAM *param,
                         uint4 *seclen, uint4 *seccnt, size_t nsections,
                         float *nword_factor)
{
  uint4 i, wfsum;
  float density_factor= 0;
  for (wfsum=0, i=0; i < nsections; i++)
  {
    if (seclen[i])
    {
      float add= (float) param->wf2[i] * seccnt[i] / seclen[i];
      wfsum+= param->wf2[i];
      density_factor+= add;
    }
  }
  if (wfsum)
  {
    density_factor= (density_factor > wfsum) ? 1 : density_factor / wfsum;
    *nword_factor= UDM_MIX_FACTOR(param->WordDensityFactor, density_factor,
                                  param->WordDensityFactorReminder,
                                  nword_factor[0]);
  }
}


static void
UdmApplyDensityFactor(UDM_SCORE_PARAM *param,
                      UDM_URL_CRD *CrdFrom,
                      size_t phr_n,
                      float *nword_factor)
{
  UDM_URL_CRD *CrdEnd, *CrdCurr;
  uint4 seclen[256];
  uint4 seccnt[256];
  size_t nsections= param->nsections;
  size_t max_secno= 0;
      
  bzero(&seclen, sizeof(uint4)*nsections);
  bzero(&seccnt, sizeof(uint4)*nsections);
  for (CrdCurr= CrdFrom, CrdEnd= CrdFrom + phr_n; CrdCurr < CrdEnd; CrdCurr++)
  {
    size_t secno= CrdCurr->secno;
    seclen[secno]= CrdCurr->seclen;
    seccnt[secno]++;
    if (max_secno < secno)
      max_secno= secno;
  }

  UdmCalcWordDensityFactor(param, seclen, seccnt, max_secno + 1, nword_factor);
}


static void
UdmGroupByURLLoop(UDM_RESULT *Res,
                  UDM_URLCRDLIST *CoordList,
                  UDM_URLSCORELIST *ScoreList,
                  UDM_SCORE_PARAM *score_param)
{
  UDM_URL_CRD *Crd= CoordList->Coords;
  UDM_URL_SCORE *CrdTo= ScoreList->Item;
  UDM_URL_CRD *CrdFrom, *CrdCurr;
  UDM_URL_CRD *CrdLast= CoordList->Coords + CoordList->ncoords;
  UDM_WIDEWORD *Res_WWList_Word= score_param->Word;
  size_t nwf_num= score_param->nwf_num;
  int search_mode= score_param->search_mode;
  size_t *count= score_param->count;
  size_t count_size= score_param->count_size;
  UDM_STACK_ITEM *temp_items= score_param->ItemList.items;
  size_t nitems= score_param->ItemList.nitems;
  char *wf2= score_param->wf2;
  char *nwf= score_param->nwf;
  float *numword_factor= score_param->numword_factor;
  float *numdistinctword_factor= score_param->numdistinctword_factor;
  int have_WordFormFactor= score_param->have_WordFormFactor;
  float WordFormFactor= score_param->WordFormFactor;
  float WordFormFactorReminder= score_param->WordFormFactorReminder;

  for (CrdFrom= Crd ; CrdFrom < CrdLast ; CrdFrom= CrdCurr)
  {
    float nword_factor;
    size_t z, phr_n, form_phr_n;
    size_t max_coord= CrdFrom->pos;
    size_t min_coord= CrdFrom->pos;
    bzero((void*)count, count_size);
    
    for (CrdCurr= CrdFrom ;
         CrdCurr < CrdLast && CrdCurr->url_id == CrdFrom->url_id;
         CrdCurr++)
    {
      size_t CrdCurr_pos= CrdCurr->pos;
      count[Res_WWList_Word[CrdCurr->num].order]++;
      if (max_coord < CrdCurr_pos)
        max_coord= CrdCurr_pos;
      if (min_coord > CrdCurr_pos)
        min_coord= CrdCurr_pos;
    }
    
    phr_n= CrdCurr - CrdFrom;
    
    switch(search_mode)
    {
      case UDM_MODE_BOOL:
        CheckPhrase(&Res->WWList, temp_items, nitems, CrdFrom, phr_n, count);
        if(!UdmCalcBoolItems(temp_items, nitems, count))
          continue;
        break;
        
      case UDM_MODE_ALL:
        if (phr_n < Res->WWList.nuniq)
          continue;
        if (Res->WWList.nuniq > 1)
        {
          for (z = 0; z < Res->WWList.nuniq; z++) if (count[z] == 0) break;
          if (z < Res->WWList.nuniq && count[z] == 0) continue;
        }
    }
    nword_factor= NUMWORD_FACTOR(numword_factor, phr_n);

    if (have_WordFormFactor)
    {
      float k;
      unsigned int phr_n2= phr_n * 2;
      for (form_phr_n= 0, z= 0 ; z < phr_n; z++)
      {
        int spell_factor[5]= {0, 2, 1, 1, 1};
        form_phr_n+= spell_factor[Res_WWList_Word[CrdFrom[z].num].origin];
      }
      k= WordFormFactor + WordFormFactorReminder * (float) form_phr_n / phr_n2;
      nword_factor*= k;
    }

    if (CrdFrom->seclen)
    {
      UdmApplyDensityFactor(score_param, CrdFrom, phr_n, &nword_factor);
    }
    

    if (search_mode != UDM_MODE_ALL)
    {
      size_t nuniq= 0;
      for (z= 0; z < Res->WWList.nuniq; z++)
      {
        if (count[z])
          nuniq++;
      }
      if (nuniq < Res->WWList.nuniq)
        nword_factor= nword_factor * numdistinctword_factor[nuniq];
    }

    if (nwf_num)
    {
      {
        size_t i;
        unsigned char secno_min= 255;
        unsigned char secno_max= 0;
        char section_vector[256];
        UDM_URL_CRD *CrdTmp;
        
        bzero((void*) &section_vector, nwf_num);
        for (CrdTmp=CrdFrom, i= phr_n; i > 0; i--, CrdTmp++)
        {
          unsigned char secno= CrdTmp->secno;
          section_vector[secno]= 1;
          if (secno_min > secno)
            secno_min= secno;
          if (secno_max < secno)
            secno_max= secno;
        }
        if (secno_min == secno_max && nwf[secno_min])
          nword_factor*= (1 -  (float) nwf[secno_min] / 16);
      }
    }
    
    CrdTo->score= UdmCalcCosineWeightQuick(score_param,
                                           CrdFrom, phr_n,
                                           wf2, nword_factor);
    CrdTo->url_id= CrdFrom->url_id;
    CrdTo++;
  }
  ScoreList->nitems= CrdTo - ScoreList->Item;
}


#if 0
static void
UdmGroupByURLLoop_OneWord_All_NoNWF(UDM_RESULT *Res,
                                    UDM_URLCRDLIST *CoordList,
                                    UDM_URLSCORELIST *ScoreList,
                                    UDM_SCORE_PARAM *score_param)
{
  UDM_URL_SCORE *CrdTo= ScoreList->Item;
  UDM_URL_CRD *CrdFrom, *CrdCurr;
  UDM_URL_CRD *CrdLast= CoordList->Coords + CoordList->ncoords;
  char *wf2= score_param->wf2;
  float *numword_factor= score_param->numword_factor;
  float max_coord_factor= score_param->max_coord_factor;
  float Rsum_factor= score_param->Rsum_factor;
  unsigned int MinCoordFactor= score_param->MinCoordFactor;
  char *added= (char*) score_param->D;
  size_t ncosine= score_param->ncosine;
  size_t weight= score_param->Word[0].weight;
  float numword_factor_1= NUMWORD_FACTOR(numword_factor, 1);
  float Rsum_factor_mul_nword_factor_1= Rsum_factor * numword_factor_1;
  unsigned int *R= score_param->R;

  bzero((void*) added, ncosine);
  
  for (CrdFrom= CoordList->Coords ; CrdFrom < CrdLast ; CrdFrom= CrdCurr)
  {
    urlid_t CrdFrom_url_id= CrdFrom->url_id;
    
    CrdCurr= CrdFrom + 1;
    if (CrdCurr >= CrdLast || CrdCurr->url_id != CrdFrom_url_id)
    {
      unsigned char secno= CrdFrom->secno;
      if (CrdFrom->seclen)
      {
        float nword_factor= numword_factor_1 *
                    MAXCOORD_FACTOR(max_coord_factor, CrdFrom->pos) *
                    MINCOORD_FACTOR(MinCoordFactor, CrdFrom->pos);

        UdmApplyDensityFactor(score_param, CrdFrom, 1, &nword_factor);
        CrdTo->score= Rsum_factor * nword_factor * (float) (R[secno]) + 0.5;
        CrdTo->url_id= CrdFrom_url_id;
        CrdTo++;
      }
      else
      {
        CrdTo->score= Rsum_factor_mul_nword_factor_1 *
                      MAXCOORD_FACTOR(max_coord_factor, CrdFrom->pos) *
                      MINCOORD_FACTOR(MinCoordFactor, CrdFrom->pos) *
                      (float) (R[secno]) + 0.5;
        CrdTo->url_id= CrdFrom_url_id;
        CrdTo++;
      }
    }
    else
    {
      unsigned short max_coord, min_coord;
      unsigned char max_secno, prev_secno;
      size_t ncoords, add, Dsum, RDsum;
      float nword_factor;

      max_coord= min_coord= CrdFrom->pos;
      max_secno= prev_secno= CrdFrom->secno;
      
      /* Add the first coord */
      add= wf2[max_secno] + weight;
      Dsum= add * add;
      RDsum= add * R[max_secno];
      added[max_secno]= 1;
      
      /* Add all other coords */
      do 
      {
        unsigned char secno= CrdCurr->secno;
        unsigned short CrdCurr_pos= CrdCurr->pos;
        if (prev_secno != secno && !added[secno])
        {
          add= wf2[secno] + weight;
          Dsum+= add * add;
          RDsum+= add * R[secno];
          added[secno]= 1;
          if (secno > max_secno)
            max_secno= secno;
        }
        if (max_coord < CrdCurr_pos)
          max_coord= CrdCurr_pos;
        else if (min_coord > CrdCurr_pos)
          min_coord= CrdCurr_pos;
        CrdCurr++;
      } while (CrdCurr < CrdLast && CrdCurr->url_id == CrdFrom_url_id);

      bzero((void*) added, max_secno + 1);
      
      ncoords= CrdCurr - CrdFrom;
      
      nword_factor= NUMWORD_FACTOR(numword_factor, ncoords) *
                    MAXCOORD_FACTOR(max_coord_factor, max_coord) *
                    MINCOORD_FACTOR(MinCoordFactor, min_coord);

      if (CrdFrom->seclen)
      {
        UdmApplyDensityFactor(score_param, CrdFrom, ncoords, &nword_factor);
      }
      
      CrdTo->score= Rsum_factor * nword_factor * (float) RDsum / sqrt(Dsum) + 0.5;
      CrdTo->url_id= CrdFrom_url_id;
      CrdTo++;
    }
  }
  ScoreList->nitems= CrdTo - ScoreList->Item;
}
#endif


static void
UdmGroupByURLLoop_OneWord_All_NoNWF2(UDM_RESULT *Res,
                                     UDM_SECTIONLIST *SectionList,
                                     UDM_URLSCORELIST *ScoreList,
                                     UDM_SCORE_PARAM *score_param)
{
  UDM_URL_SCORE *CrdTo= ScoreList->Item;
  UDM_SECTION *From;
  UDM_SECTION *Last= SectionList->Section + SectionList->nsections;
  char *wf2= score_param->wf2;
  float *numword_factor= score_param->numword_factor;
  float max_coord_factor= score_param->max_coord_factor;
  float Rsum_factor= score_param->Rsum_factor;
  unsigned int MinCoordFactor= score_param->MinCoordFactor;
  char *added= (char*) score_param->D;
  size_t ncosine= score_param->ncosine;
  size_t weight= score_param->Word[0].weight;
  /*
  TODO:
  float numword_factor_1= NUMWORD_FACTOR(numword_factor, 1);
  float Rsum_factor_mul_nword_factor_1= Rsum_factor * numword_factor_1;
  */
  unsigned int *R= score_param->R;
  
  bzero((void*) added, ncosine);
  
  for (From= SectionList->Section ; From < Last ; )
  {
    urlid_t From_url_id= From->url_id;
    uint4 min_pos= From->minpos;
    uint4 max_pos= From->maxpos;
    size_t ncoords= 0;
    unsigned char max_secno= 0;
    size_t Dsum= 0;
    size_t RDsum= 0;
    uint4 wfsum= 0;
    float density_factor= 0;
    unsigned char prev_secno= From->secno;
    float min_max_factor= 1;
    
    /* TODO: one coord, for nsections==1 and ncoords=1 */
    
    for (; ;)
    {
      unsigned char secno= From->secno;
      ncoords+= From->ncoords;
      
      if (prev_secno != secno)
      {
        min_max_factor*= MAXCOORD_FACTOR(max_coord_factor, max_pos) *
                         MINCOORD_FACTOR(MinCoordFactor, min_pos);
        min_pos= From->minpos;
        max_pos= From->maxpos;
        prev_secno= secno;
      }
      else
      {
        if (max_secno < secno)
          max_secno= secno;
        if (min_pos > From->minpos)
          min_pos= From->minpos;
        if (max_pos < From->maxpos)
          max_pos= From->maxpos;
      }
      
      if (From->seclen)
      {
        float add= (float) wf2[secno] * From->ncoords / From->seclen;
        wfsum+= wf2[secno];
        density_factor+= add;
      }
      
      if (!added[secno])
      {
        uint4 add= wf2[secno] + weight;
        Dsum+= add * add;
        RDsum+= add * R[secno];
        added[secno]= 1;
        if (max_secno < secno)
          max_secno= secno;
      }
      
      From++;
      
      if (From >= Last || From->url_id != From_url_id)
      {
        float nword_factor;
        min_max_factor*= MAXCOORD_FACTOR(max_coord_factor, max_pos) *
                         MINCOORD_FACTOR(MinCoordFactor, min_pos);
        /*WordFormFactor*/
        density_factor= (density_factor > wfsum) ? 1 : density_factor / wfsum;
        nword_factor= NUMWORD_FACTOR(numword_factor, ncoords);
        if (wfsum)
          nword_factor= UDM_MIX_FACTOR(score_param->WordDensityFactor,
                                       density_factor,
                                       score_param->WordDensityFactorReminder,
                                       nword_factor);
        CrdTo->score= Rsum_factor * nword_factor * min_max_factor *
                      (float) RDsum / sqrt(Dsum) + 0.5;
        CrdTo->url_id= From_url_id;
        CrdTo++;
        
        if (max_secno)
          bzero((void*)added, (size_t) max_secno + 1);

        break;
      }
    }
  }
  ScoreList->nitems= CrdTo - ScoreList->Item;
}


static void
UdmScoreParamInit(UDM_SCORE_PARAM *prm,
                  UDM_AGENT *query,
                  UDM_DB *db,
                  UDM_RESULT *Res)
{
  size_t i;
  float DefWordDensityFactor;
  UDM_VARLIST *Vars= &query->Conf->Vars;
  bzero((void*) prm, sizeof(UDM_SCORE_PARAM));
  prm->nsections = UdmVarListFindInt(Vars, "NumSections", 256);
  prm->dst_offs= Res->WWList.nuniq * prm->nsections;
  prm->nwf_offs= Res->WWList.nuniq * prm->nsections + 1;
  prm->max_coord_factor= ((float)UdmVarListFindInt(Vars, "MaxCoordFactor", 255)) / 0xFFFFFF;
  prm->MinCoordFactor= UdmVarListFindInt(Vars, "MinCoordFactor", 0);
  prm->have_WordFormFactor= UdmVarListFindInt(Vars, "WordFormFactor", 255) != 255;
  prm->WordFormFactor= ((float)UdmVarListFindDouble(Vars, "WordFormFactor", 255)) / 255;
  prm->WordFormFactorReminder= 1 - prm->WordFormFactor;
  prm->SaveSectionSize= UdmVarListFindBool(Vars, "SaveSectionSize", 1);
  DefWordDensityFactor= prm->SaveSectionSize ? 25 : 0;
  prm->WordDensityFactor= ((float)UdmVarListFindDouble(Vars, "WordDensityFactor", DefWordDensityFactor)) / 256;
  prm->WordDensityFactorReminder= 1 - prm->WordDensityFactor;
  prm->dst_weight= (unsigned int) UdmVarListFindInt(Vars, "WordDistanceWeight", 255);
  UdmWeightFactorsInit2(prm->wf, Vars, &db->Vars, "wf");
  prm->nwf_num= UdmWeightFactorsInit2(prm->nwf, Vars, &db->Vars, "nwf");

  for (i= 0; i < 256; i++)
    prm->wf2[i]= prm->wf[i] << 2;

  prm->ncosine= Res->WWList.nuniq * prm->nsections + 1;
  prm->D_size= prm->ncosine * sizeof(unsigned int);
}


static void
UdmNumDistinctWordFactorInit(unsigned int numdistinctwordfactor,
                             float *numdistinctword_factor, size_t nuniq)
{
  size_t i;
  for (i= 0; i < nuniq; i++)
  {
    unsigned int k= numdistinctwordfactor;
    float x= (float) i / nuniq;
    numdistinctword_factor[i]= ((x * x * x * x * k) + (256-k)) / 256;
  }
}


/*
  Initializing ncoords and search_mode dependent member,
  and members requiring allocation
*/
static int
UdmScoreParamInitStep2(UDM_SCORE_PARAM *score_param,
                       UDM_AGENT *query, UDM_RESULT *Res,
                       size_t ncoords, int search_mode)
{
  size_t  i, Rsum;
  UDM_WIDEWORD *Res_WWList_Word= Res->WWList.Word;

  float numwordfactor= ((float)UdmVarListFindDouble(&query->Conf->Vars,
                                                "NumWordFactor", 25.5)) / 255;
  score_param->NumDistinctWordFactor= (unsigned int) UdmVarListFindInt(
                                                      &query->Conf->Vars,
                                                     "NumDistinctWordFactor", 0);
  UdmNumWordFactorInit(numwordfactor, score_param->numword_factor, ncoords);
  UdmNumDistinctWordFactorInit(score_param->NumDistinctWordFactor,
                               score_param->numdistinctword_factor,
                               Res->WWList.nuniq);  
  score_param->count_size= Res->WWList.nuniq * sizeof(size_t);
  score_param->count= (size_t*)UdmMalloc(score_param->count_size);
  score_param->R= (unsigned int*)UdmMalloc(score_param->D_size);
  score_param->D= (unsigned int*)UdmMalloc(score_param->D_size);
  if (!score_param->count || !score_param->R || !score_param->D)
    return UDM_ERROR;

  bzero((void*) score_param->R, score_param->D_size);
  
  for(Rsum=0, i= 0; i < Res->WWList.nwords; i++)
  {
    size_t secno;
    if (Res_WWList_Word[i].origin != UDM_WORD_ORIGIN_QUERY)
      continue;
    for (secno= 0; secno < score_param->nsections; secno++)
    {
      size_t offs= Res_WWList_Word[i].order * score_param->nsections + secno;
      score_param->R[offs] = score_param->wf2[secno] + Res_WWList_Word[i].weight;
      Rsum+= score_param->R[offs] * score_param->R[offs];
    }
  }
  
  Rsum+= score_param->R[score_param->dst_offs] * score_param->R[score_param->dst_offs];
  score_param->Rsum_factor= 100000 / sqrt(Rsum);
  
  if (Res->ItemList.ncmds > 0 || search_mode == UDM_MODE_BOOL)
  {
    if (UDM_OK != UdmStackItemListCopy(&score_param->ItemList,
                                       &Res->ItemList, search_mode))
      return UDM_ERROR;
    search_mode= UDM_MODE_BOOL;
  }

  score_param->Word= Res->WWList.Word;
  score_param->nuniq= Res->WWList.nuniq;
  score_param->nwords= Res->WWList.nwords;
  score_param->search_mode= search_mode;
  return UDM_OK;
}


static void
UdmScoreParamFreeStep2(UDM_SCORE_PARAM *score_param)
{
  UdmStackItemListFree(&score_param->ItemList);
  UDM_FREE(score_param->D);
  UDM_FREE(score_param->R);
  UDM_FREE(score_param->count);
}


static void
UdmGroupByURLInternal(UDM_AGENT *query,
                      UDM_RESULT *Res,
                      UDM_URLCRDLIST *CoordList,
                      UDM_URLSCORELIST *ScoreList,
                      UDM_SCORE_PARAM *score_param,
                      int search_mode)
{
  if(!CoordList->ncoords) return;

  if (UdmScoreParamInitStep2(score_param, query, Res, CoordList->ncoords, search_mode))
    goto err;

#if 0
  if (search_mode == UDM_MODE_ALL &&
      Res->WWList.nwords == 1 &&
      score_param->nwf_num == 0 &&
      !numdistinctwordfactor)
  {
    UdmGroupByURLLoop_OneWord_All_NoNWF(Res, CoordList, ScoreList, score_param);
  }
  else
#endif
  {
    UdmGroupByURLLoop(Res, CoordList, ScoreList, score_param);
  }

err:

  UdmScoreParamFreeStep2(score_param);
  return;
}



static void
UdmGroupByURLInternal2(UDM_AGENT *query,
                       UDM_RESULT *Res,
                       UDM_SECTIONLIST *SectionList,
                       UDM_URLSCORELIST *ScoreList,
                       UDM_SCORE_PARAM *score_param,
                       int search_mode)
{
  size_t i, ncoords;
  for (ncoords=0, i= 0; i < SectionList->nsections; i++)
    ncoords+= SectionList->Section[i].ncoords;

  if(!ncoords) return;

  if (UdmScoreParamInitStep2(score_param, query, Res, ncoords, search_mode))
    goto err;

  if (search_mode == UDM_MODE_ALL &&
      Res->WWList.nwords == 1 &&
      score_param->nwf_num == 0 &&
      !score_param->NumDistinctWordFactor)
  {
    UdmGroupByURLLoop_OneWord_All_NoNWF2(Res, SectionList, ScoreList, score_param);
  }
  else
  {
    /*UdmGroupByURLLoop(Res, CoordList, ScoreList, score_param);*/
  }

err:

  UdmScoreParamFreeStep2(score_param);
  return;
}


void UdmGroupByURL2(UDM_AGENT *query,
                   UDM_DB *db,
                   UDM_RESULT *Res,
                   UDM_SECTIONLIST *SectionList,
                   UDM_URLSCORELIST *ScoreList)
{
  UDM_SCORE_PARAM *prm;
  int search_mode= UdmSearchMode(UdmVarListFindStr(&query->Conf->Vars, "m", "all"));
  size_t threshold= UdmVarListFindInt(&query->Conf->Vars, "StrictModeThreshold", 0);
  size_t ncoords= (search_mode == UDM_MODE_ALL) && threshold ?
                  SectionList->nsections : 0;
  size_t nbytes;

  UdmWideWordListSetOriginWeight(&Res->WWList);
  
  if (!(prm= UdmMalloc(sizeof(UDM_SCORE_PARAM))))
    return;
  
  UdmScoreParamInit(prm, query, db, Res);


  nbytes= SectionList->nsections * sizeof(UDM_URL_SCORE);
  ScoreList->Item= (UDM_URL_SCORE*) UdmMalloc(nbytes);
  UdmGroupByURLInternal2(query, Res, SectionList, ScoreList, prm, search_mode);
  if (ncoords && (ScoreList->nitems < threshold))
  {
    size_t strict_mode_found= ScoreList->nitems;
    UdmLog(query, UDM_LOG_DEBUG,
          "Too few results: %d, Threshold: %d, group in ANY mode", 
          strict_mode_found, threshold);
    UdmGroupByURLInternal2(query, Res, SectionList, ScoreList, prm, UDM_MODE_ANY);
    if (ScoreList->nitems > strict_mode_found)
      UdmVarListReplaceInt(&query->Conf->Vars, "StrictModeFound", strict_mode_found);
  }
 
  UdmFree(prm);
}


void UdmGroupByURL(UDM_AGENT *query,
                   UDM_DB *db,
                   UDM_RESULT *Res,
                   UDM_URLCRDLIST *CoordList,
                   UDM_URLSCORELIST *ScoreList)
{
  UDM_SCORE_PARAM *prm;
  int search_mode= UdmSearchMode(UdmVarListFindStr(&query->Conf->Vars, "m", "all"));
  size_t threshold= UdmVarListFindInt(&query->Conf->Vars, "StrictModeThreshold", 0);
  size_t ncoords= (search_mode == UDM_MODE_ALL) && threshold ?
                  CoordList->ncoords : 0;
  size_t nbytes;

  UdmWideWordListSetOriginWeight(&Res->WWList);
  
  if (!(prm= UdmMalloc(sizeof(UDM_SCORE_PARAM))))
    return;
  
  UdmScoreParamInit(prm, query, db, Res);

  nbytes= CoordList->ncoords * sizeof(UDM_URL_SCORE);
  ScoreList->Item= (UDM_URL_SCORE*) UdmMalloc(nbytes);
  UdmGroupByURLInternal(query, Res, CoordList, ScoreList, prm, search_mode);
  
  if (ncoords && (ScoreList->nitems < threshold))
  {
    size_t strict_mode_found= ScoreList->nitems;
    UdmLog(query, UDM_LOG_DEBUG,
          "Too few results: %d, Threshold: %d, group in ANY mode", 
          strict_mode_found, threshold);
    UdmGroupByURLInternal(query, Res, CoordList, ScoreList, prm, UDM_MODE_ANY);
    if (ScoreList->nitems > strict_mode_found)
      UdmVarListReplaceInt(&query->Conf->Vars, "StrictModeFound", strict_mode_found);
  }
  
  UdmFree(prm);
}
